-- $Id: objects.lua 3333 2018-04-20 00:32:38Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2018 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

local Entity = require 'engine.Entity'

newTalentType {
  name = 'Object powers',
  type = 'misc/object',
  description = [[Powers conferred by objects]],
}

newTalent {
  name = 'Dig',
  short_name = 'OBJ_DIG',
  display_entity = Entity.new { image='talents/dig.png' },
  type = 'misc/object',
  mode = 'activated',
  points = 0,
  repeat_action = function(self)
    local dig = self.resting.check_info
    if self.adj_digging > dig.skill_req + rng.range(1, dig.hardness) then
      game.logPlayer(self, dig.finish)
      self.resting.dig_success = true
      return false, nil
    end
    if rng.chance(2) then
      if self.adj_digging < dig.skill_req then
	game.logPlayer(self, 'You fail to make even the slightest of progress.')
	return false, 'terrain is too hard'
      elseif self.adj_digging < dig.skill_req_1pct then
	game.logPlayer(self, 'This will take some time.')
      end
    end
    return true, nil
  end,
  action = function(self, t)
    -- This is a rough mix of T4's Dig talent and do_cmd_tunnel_test(),
    -- do_cmd_tunnel_aux() and do_cmd_tunnel() in src/cmd2.c.
    local tg = { range=1, type='bolt', nolock=true }
    local x, y = self:getTarget(tg)
    if not x or not y then return nil end

    local a = game.level.map(x, y, Map.ACTOR)
    if a then
      game.logPlayer(self, 'There is a monster in the way!')
      self:bumpInto(a)
      return true
    end
    if not game.level.map.seens(x, y) then
      game.logPlayer(self, 'You cannot see to tunnel there.')
      return nil
    end
    local g = game.level.map(x, y, Map.TERRAIN)
    if not g then
      game.logPlayer(self, 'There is nothing to dig there. (?)')
      return nil
    end
    if g:floor_bold() then
      game.logPlayer(self, 'There is nothing to dig there.')
      return nil
    end
    if not g:hasFlags('TUNNELABLE') or not g.dig then
      game.logPlayer(self, 'You cannot dig through that.')
      return nil
    end

    local wait = function()
      local co = coroutine.running()
      local ok = false
      self.repeat_action = t.repeat_action
      self.rest_info = g.dig
      self:restInit(99, 'digging', 'dug', function(cnt, max)
	ok = self.resting.dig_success
	coroutine.resume(co)
      end)
      coroutine.yield()
      if not ok then return false end
      return true
    end
    if wait() then
      local proj = spell.set_project { grid=true }
      self:project(tg, x, y, DamageType.DIG, 1)
      spell.setProject(proj)
      game.player:playerFOV()
    end

    return true
  end,
  info = function(self, t)
    local fmt = [[Dig through a wall or other terrain feature using your wielded digging tool (current dig power %d).]]
    return fmt:format(self.adj_digging)
  end,
}

newTalent {
  name = 'Swap Position',
  short_name = 'OBJ_SWAP_POSITION',
  display_entity = Entity.new { image='talents/swap_position.png' },
  type = 'misc/object',
  mode = 'activated',
  power = { level=15, stat='DEX', diff=16 },
  mana = 12,
  range = 20,
  points = 0,
  action = function(self, t)
    if self:has('ANCHOR') then
      game.logPlayer(self, [[The space-time continuum can't be disrupted.]])
      return false
    end
    local tg = {
      type = 'hit',
      default_target = self,
      range = self:getTalentRange(t),
      pass_terrain = true,
      requires_knowledge = false,
    }
    local tx, ty, target = self:getTarget(tg)
    if not target then
      game.logPlayer(self, [[There is nothing there to swap with.]])
      return false
    end
    local DamageType = require 'engine.DamageType'
    if target:resistDam(DamageType.TELEPORT) then
      game.logPlayer(self, [[Your teleportation is blocked!]])
      return false
    end
    if not self:canMove(tx, ty, true) or not target:canMove(self.x, self.y, true) then
      game.logPlayer(self, [[Your teleportation is blocked!]])
      return false
    end
    local sx, sy = self.x, self.y
    self.x, self.y = nil, nil
    self:move(tx, ty, true)
    target.x, target.y = nil, nil
    target:move(sx, sy, true)
    return true
  end,
  info = function(self, t)
    return [[Swap position with a targeted creature.]]
  end,
}

newTalent {
  name = 'Change the World',
  short_name = 'OBJ_CHANGE_THE_WORLD',
  display_entity = Entity.new { image='talents/change_the_world.png' },
  type = 'misc/object',
  mode = 'activated',
  power = { level=1, stat='CHA', diff=6 },
  mana = 30,
  points = 0,
  action = function(self, t)
    game.log('Sorry, not implemented (OBJ_CHANGE_THE_WORLD)')
  end,
  info = 'placeholder',
}

newTalent {
  name = 'Panic Hit',
  short_name = 'OBJ_PANIC_HIT',
  display_entity = Entity.new { image='talents/panic_hit.png' },
  type = 'misc/object',
  mode = 'activated',
  power = { level=10, stat='DEX', diff=14 },
  mana = 12,
  points = 0,
  action = function(self, t)
    local Map = require 'mod.class.Map'
    local function match(x, y)
      -- Hostile actor in grid.
      local a = game.level.map(x, y, Map.ACTOR)
      return a and a:reactionToward(self) < 0
    end
    local tx, ty = util.select_adjacent(self, match, 'There is nothing here to attack.', 'There is nothing there to attack.')
    if not tx or not ty then return end
    local target = game.level.map(tx, ty, Map.ACTOR)
    self:attackTarget(target)
    spell.teleport(self, 30)
    return true
  end,
  info = function(self, t)
    return [[Attack an adjacent hostile creature in melee and then immediately teleport away (range 30).]]
  end,
}

newTalent {
  name = 'Weigh Magic',
  short_name = 'OBJ_WEIGH_MAGIC',
  display_entity = Entity.new { image='talents/weight-magic.png' },
  type = 'misc/object',
  mode = 'activated',
  power = { level=6, stat='INT', diff=10 },
  mana = 6,
  points = 0,
  action = function(self, t)
    game.log('Sorry, not implemented (OBJ_WEIGH_MAGIC)')
  end,
  info = 'placeholder',
}

newTalent {
  name = 'Magic Map',
  short_name = 'OBJ_MAGIC_MAP',
  display_entity = Entity.new { image='talents/magic_map.png' },
  type = 'misc/object',
  mode = 'activated',
  power = { level=7, stat='WIS', diff=15 },
  mana = 10,
  points = 0,
  action = function(self, t)
    self:computeFOV(40, 'block_sense', spell.map_area_callback, true, true, true)
    return true
  end,
  info = function(self, t)
    return [[Maps the area around you.]]
  end,
}

newTalent {
  name = 'Teleport',
  short_name = 'OBJ_TELEPORT',
  display_entity = Entity.new { image='talents/teleport.png' },
  type = 'misc/object',
  mode = 'activated',
  power = { level=7, stat='WIS', diff=15 },
  mana = 7,
  points = 0,
  action = function(self, t)
    spell.teleport(self, 10 + 4*(self.level or 0))
    return true
  end,
  info = function(self, t)
    local range = 10 + 4*(self.level or 0)
    local fmt = [[Teleport randomly within range %d (range increase with caster level).]]
    return fmt:format(range)
  end,
}

newTalent {
  name = 'Restore Life',
  short_name = 'OBJ_RESTORE_LIFE',
  display_entity = Entity.new { image='talents/restore_life.png' },
  type = 'misc/object',
  mode = 'activated',
  power = { level=30, stat='WIS', diff=18 },
  mana = 30,
  points = 0,
  action = function(self, t)
    game.logPlayer(self, 'You attempt to restore your lost energies.')
    if self.restoreExp then self:restoreExp() end
    return true
  end,
  info = function(self, t)
    return [[Restores lost experience.]]
  end,
}

newTalent {
  name = 'Midas Touch',
  short_name = 'OBJ_MIDAS_TOUCH',
  display_entity = Entity.new { image='talents/midas_touch.png' },
  type = 'misc/object',
  mode = 'activated',
  power = { level=10, stat='INT', diff=12 },
  mana = 5,
  points = 0,
  action = function(self, t)
    game.log('Sorry, not implemented (OBJ_MIDAS_TOUCH)')
  end,
  info = 'placeholder',
}

newTalent {
  name = 'Illuminate',
  short_name = 'OBJ_ILLUMINATE',
  display_entity = Entity.new { image='talents/illuminate.png' },
  type = 'misc/object',
  mode = 'activated',
  power = { level=3, stat='INT', diff=10 },
  mana = 2,
  points = 0,
  action = function(self, t)
    local ds = math.min(1, math.floor((self.level or 0)/2))
    local dam = rng.dice(2, ds)
    local r = 1 + math.floor((self.level or 0)/10)
    game.logPlayer(self, 'You are surrounded by a bright light.')
    spell.ball_spell(who, nil, DamageType.LITE_WEAK, rng.dice(2, ds), r, who.x, who.y, 0)
    spell.lite_room(self.x, self.y)
    return true
  end,
  info = function(self, t)
    local ds = math.min(1, math.floor((self.level or 0)/2))
    local r = 1 + math.floor((self.level or 0)/10)
    local lf = util.colorize_dam(DamageType.LITE_WEAK)
    local fmt = [[Permanently illuminate your immediate surroundings, doing 2d%d %s damage to light-sensitive creatures in radius %d (damage and radius increase with caster level).]]
    return fmt:format(lf, ds, r)
  end,
}

newTalent {
  name = 'Stone to Mud',
  short_name = 'OBJ_STONE_TO_MUD',
  display_entity = Entity.new { image='talents/stone_to_mud.png' },
  type = 'misc/object',
  mode = 'activated',
  power = { level=20, stat='STR', diff=12 },
  mana = 10,
  points = 0,
  action = function(self, t)
    game.logPlayer(self, 'You bash at a stone wall.')
    return spell.beam_spell(self, t, DamageType.KILL_WALL, 20 + rng.range(1, 30))
  end,
  info = function(self, t)
    return [[Destroy a targeted wall.]]
  end,
}

newTalent {
  name = [[Earthquake]],
  short_name = 'OBJ_EARTHQUAKE',
  display_entity = Entity.new { image='talents/earthquake.png' },
  type = 'misc/object',
  mode = 'activated',
  power = { leve=12, stat='STR', diff=16 },
  mana = 12,
  points = 0,
  action = function(self, t)
    spell.earthquale(self.x, self.y, 10)
    return true
  end,
  info = function(self, t)
    return [[Triggers an earthquake of radius 10 centered on the caster.]]
  end,
}

newTalent {
  name = [[Mind Blast]],
  short_name = 'OBJ_MIND_BLAST',
  display_entity = Entity.new { image='talents/mind_blast.png' },
  type = 'misc/object',
  mode = 'activated',
  power = { leve=5, stat='WIS', diff=15 },
  mana = 3,
  points = 0,
  range = 18,
  action = function(self, t)
    game.logPlayer(self, 'You concentrate...')
    local dd = 3 + math.floor((self.level - 1)/5)
    return spell.bolt_spell(self, t, DamageType.PSI, rng.dice(dd, 3))
  end,
  info = function(self, t)
    local dd = 3 + math.floor((self.level - 1)/5)
    local pf = util.colorize_damtype(DamageType.PSI)
    local ret = [[Focuses mental energy into a powerful bolt doing %dd3 %s damage.  Damage increases with caster level.]]
    return ret:format(dd, pf)
  end,
}

newTalent {
  name = [[Telekinesis]],
  short_name = 'OBJ_TELEKINESIS',
  display_entity = Entity.new { image='talents/telekinesis.png' },
  type = 'misc/object',
  mode = 'activated',
  power = { leve=9, stat='WIS', diff=14 },
  mana = 9,
  points = 0,
  action = function(self, t)
    game.log('Sorry, not implemented (OBJ_TELEKINESIS)')
  end,
  info = 'placeholder',
}

newTalent {
  name = [[Cold Touch]],
  short_name = 'OBJ_COLD_TOUCH',
  display_entity = Entity.new { image='talents/cold_touch.png' },
  type = 'misc/object',
  mode = 'activated',
  power = { leve=2, stat='CON', diff=11 },
  mana = 2,
  points = 0,
  range = 1,
  action = function(self, t)
    return spell.bolt_spell(self, t, DamageType.COLD, 2*self.level)
  end,
  info = function(self, t)
    local cf = util.colorize_damtype(DamageType.COLD)
    local ret = [[Touch an adjacent creature to do %d %s damage.  Damage increases with caster level.]]
    return ret:format(2*self.level, cf)
  end,
}

newTalent {
  name = [[Dazzle]],
  short_name = 'OBJ_DAZZLE',
  display_entity = Entity.new { image='talents/dazzle.png' },
  type = 'misc/object',
  mode = 'activated',
  power = { leve=7, stat='CHR', diff=8 },
  mana = 15,
  points = 0,
  action = function(self, t)
    local pow = self.level*4
    spell.los_spell(self, t, DamageType.STUN, pow)
    spell.los_spell(self, t, DamageType.OLD_CONF, pow)
    spell.los_spell(self, t, DamageType.TURN_ALL, pow)
    return true
  end,
  info = function(self, t)
    local ret = [[Emit a dazzling burst of bright light.  All creatures in line of sight will be stunned, confused and terrified (with power %d; power increases with caster level).]]
    return ret:format(self.level*4)
  end,
}
