-- $Id: pray.lua 3494 2018-11-20 00:45:20Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2017 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

local Entity = require 'engine.Entity'

newTalentType {
  name = 'Prayers',
  type = 'player/prayer',
  description = [[Praying to one's god]],
}

local function piety_mod(who)
  if not who.piety_regen_f then return '' end
  local no_pray = who:piety_regen_f(false)
  local yes_pray = who:piety_regen_f(true)
  if no_pray == yes_pray then return '' end
  local cur_pray = who:piety_regen_f()
  local ret = '· %s piety regen (currently %s every %d turns)'
  return ret:format(util.colorize_num(yes_pray - no_pray),
		    util.colorize_num(cur_pray),
		    math.floor(who.piety_regen_step/10))
end

local function info_message(who, t)
  return 'Pray to ' .. t.god_name .. '.  This will have the following effects at your current piety:\n\n' .. t.describe(who, t)
end

local function sustain_message(who, t, p)
  local s = 'You are praying to ' .. t.god_name .. ', with the following effects:\n\n' .. t.describe(who, t)
  return s:toTString()
end

local function use_message(who, t)
  -- NOTE:  This is called from ActorTalents:postUseTalent(), which is
  -- called before the sustain is actually activated or deactivated.
  local which = who:isTalentActive(t.id) and 'stops' or 'begins'
  return ('@Source@ %s praying to %s.'):format(which, t.god_name)
end

newTalent {
  name = 'Pray to Eru',
  short_name = 'PRAYER_ERU',
  display_entity = Entity.new { image='talents/pray.png' },
  god_name = 'Eru Ilúvatar',
  sust_display = function(p) return 'Praying (Eru)' end,
  type = 'player/prayer',
  mode = 'sustained',
  info = info_message,
  sustain_describe = sustain_message,
  message = use_message,
  activate = function(who, def) return {} end,
  deactivate = function(who, def, p) return true end,
  describe = function(who, t)
    local ret = piety_mod(who)
    local c = math.min(who.piety, 50000) - (who.level or 0)*300
    c = math.max(math.floor(c/1000), 0)
    ret = ret .. ('\n· %d%% chance to block a melee attack from a level-%d evil monster'):format(c, who.level)
    if who.piety > 100000 then
      ret = ret .. '\n· 70% chance to be resurrected upon death'
    end
    return ret
  end
}

newTalent {
  name = 'Pray to Manwë',
  short_name = 'PRAYER_MANWE',
  display_entity = Entity.new { image='talents/pray.png' },
  god_name = 'Manwë Súlimo',
  sust_display = function(p) return 'Praying (Manwë)' end,
  type = 'player/prayer',
  mode = 'sustained',
  info = info_message,
  sustain_describe = sustain_message,
  message = use_message,
  activate = function(who, def) return {} end,
  deactivate = function(who, def, p) return true end,
  describe = function(who, t)
    local ret = piety_mod(who)
    local p = who.piety
    ret = ret .. util.colorize_num(math.min(math.floor(p/5000), 7), '\n· %+d speed')
    if p >= 7000 then
      ret = ret .. '\n· Free action'
      if p >= 15000 then ret = ret .. '\n· Flight' end
    end
    return ret
  end
}

newTalent {
  name = 'Pray to Tulkas',
  short_name = 'PRAYER_TULKAS',
  display_entity = Entity.new { image='talents/pray.png' },
  god_name = 'Tulkas',
  sust_display = function(p) return 'Praying (Tulkas)' end,
  type = 'player/prayer',
  mode = 'sustained',
  info = info_message,
  sustain_describe = sustain_message,
  message = use_message,
  activate = function(who, def) return {} end,
  deactivate = function(who, def, p) return true end,
  active_callbacks = {
    on_wpn_hit_dam_mod = function(who, t, target, wpn, p)
      -- Doesn't apply to unarmed "weapon" hits.
      if wpn.type == 'weapon' then
	local dd = (wpn.to_d or 0) + (who.to_d_melee or 0)
	if rng.percent(who:getStat('WIS', {scale=130}) - (target.level or 0)) and who.piety > 1000 then
	  game.logPlayer(who, 'You feel the hand of Tulkas helping your blow.')
	  p.bonus = p.bonus + dd*2
	else
	  p.bonus = p.bonus + dd
	end
	-- Slight Hack(TM):  In the original T2 code, this bonus is applied
	-- instead of the weapon's to_d bonus if we're praying.  The caller
	-- has already added the weapon's to_d bonus, though, so we
	-- subtract it back out here.
	p.bonus = p.bonus - (wpn.to_d or 0)
      end
    end,
  },
  describe = function(who, t)
    local ret = piety_mod(who)
    local inv = who:getInven('MAIN_HAND')
    if inv and #inv > 0 then
      local dd = (inv[1].to_d or 0) + (who.to_d_melee or 0)
      local lev = who.level or 0
      local c = who:getStat('WIS', {scale=130}) - lev
      c = util.bound(c, 0, 100)
      local f = [[· %+d damage with your current weapon
· %d%% chance (vs. level-%d monster) of additional %+d damage bonus]]
      ret = ret ..'\n' .. f:format(dd, c, lev, dd)
    end
    return ret
  end
}

newTalent {
  name = 'Pray to Melkor',
  short_name = 'PRAYER_MELKOR',
  display_entity = Entity.new { image='talents/pray.png' },
  god_name = 'Melkor Bauglir',
  sust_display = function(p) return 'Praying (Melkor)' end,
  type = 'player/prayer',
  mode = 'sustained',
  info = info_message,
  sustain_describe = sustain_message,
  message = use_message,
  activate = function(who, def) return {} end,
  deactivate = function(who, def, p) return true end,
  -- Probability of auto-casting Curse (cf. src/cmd1.c line 2828)
  curse_chance = function(who, t, target)
    local lev = spell.get_spell_level(who, who.T_MELKOR_CURSE, 100)
    if lev < 10 then return 0 end
    local wis = who:getStat('WIS', {scale=30})
    local pr = math.floor((wis * lev)/target.level)
    return util.bound(pr, 1, 100)
  end,
  active_callbacks = {
    on_melee_blow = function(who, t, target, hit, wpn)
      -- Doesn't apply to unarmed "weapon" hits.
      if wpn.type == 'weapon' then
	-- Melkor Curse spell (cf. src/cmd1.c line 2828)
	local pr = t.curse_chance(who, t, target)
	if pr > 0 and who.piety > 5000 and rng.percent(pr) then
	  who:callTalent(who.T_MELKOR_CURSE, 'do_curse', target)
	end
      end
    end,
  },
  describe = function(who, t)
    local ret = piety_mod(who)
    if who.piety > 5000 then
      ret = ret .. '\n· Invisibility'
      local prob = t.curse_chance(who, t, {level=who.level})
      if prob > 0 then
	local fmt = '\n· %d%% chance (vs. level-%d opponent) of auto-casting Curse with no piety cost on each weapon hit (increases with Wisdom and spell level)'
	ret = ret .. fmt:format(prob, who.level)
      end
    end
    if who.piety > 5500 then
      local prob = math.floor(who.piety/500) - 10
      local n = math.max(1, math.floor(who.piety/5000))
      local fmt = '\n· %d%% chance of summoning up to %d friendly undead or demons when you take damage reducing you to below 25%% max life (chance and number of summons increase with piety)'
      ret = ret .. fmt:format(prob, n)
    end
    if who.piety > 15000 then
      local DamageType = require 'engine.DamageType'
      local dn = util.colorize_damtype(DamageType.FIRE)
      ret = ret .. '\n· #{bold}#Immune to:#{normal}# '..dn
    end
    return ret
  end
}

newTalent {
  name = 'Pray to Yavanna',
  short_name = 'PRAYER_YAVANNA',
  display_entity = Entity.new { image='talents/pray.png' },
  god_name = 'Yavanna Kementári',
  sust_display = function(p) return 'Praying (Yavanna)' end,
  type = 'player/prayer',
  mode = 'sustained',
  info = info_message,
  sustain_describe = sustain_message,
  message = use_message,
  activate = function(who, def) return {} end,
  deactivate = function(who, def, p) return true end,
  describe = function(who, t)
    local ret = piety_mod(who)
    if who.piety >= 9000 then ret = ret .. '\n· Tree walking' end
    return ret
  end
}
