-- $Id: PlasmaWilderness.lua 3088 2017-09-23 19:58:29Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2017 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require 'engine.class'
local Static = require 'engine.generator.map.Static'
local Zone = require 'mod.class.Zone'

module(..., package.seeall, class.inherit(Static))

-- Mostly duplicated from engine.generator.map.Static:init(), with small
-- tweaks.
function _M:init(zone, map, level, data)
  engine.Generator.init(self, zone, map, level)
  self.grid_list = zone.grid_list
  self.subgen = {}
  self.spots = {}
  data.static_replace_tiles = data.static_replace_tiles or {}
  self.data = data
  data.__import_offset_x = data.__import_offset_x or 0
  data.__import_offset_y = data.__import_offset_y or 0

  local q_id, q = game.state:getWildernessQuest()
  if q_id and q and not q:isEnded() then
    if q.merge_map_gen then
      self.data = table.merge(table.clone(data, true), q.merge_map_gen, true)
    end
  end

  if data.adjust_level then
    self.adjust_level = {
      base = zone.base_level,
      lev = self.level.level,
      min = data.adjust_level[1],
      max = data.adjust_level[2],
    }
  else
    self.adjust_level = {
      base = zone.base_level,
      lev = self.level.level,
      min = 0,
      max = 0,
    }
  end

  if data.map then self:loadMap(data.map) end
end

function _M:generate(lev, old_lev)
  local Map = require 'mod.class.Map'

  -- Generate wilderness terrain, and overlay the static map if specified.
  self:genTerrain()
  local ux = math.floor(self.level.map.w/2)
  local uy = math.floor(self.level.map.h/2)
  local dx, dy, spots = ux, uy, {}
  if self.gen_map then
    ux, uy, dx, dy, spots = Static.generate(self, lev, old_lev)
  end

  local mx, my = game.player.wild_x, game.player.wild_y
  local g = game.state:getWorldGrid(mx, my)
  if g.wild_down then
    -- Use a fixed RNG seed so the stairs will always be in the same place.
    rng.seed(game.state:getWorldMapSeed(mx, my))
    local mw, mh = self.level.map.w, self.level.map.h
    dx, dy = rng.range(6, mw-6), rng.range(6, mh-6)
    ux, uy = dx, dy
    self.level.map(dx, dy, Map.TERRAIN, self.grid_list[g.wild_down])
    self.level.map.remembers(dx, dy)
    self.level.map.lites(dx, dy)
    -- Done with fixed-RNG stuff; go back to random.
    rng.seed(-1)
  end

  self:borderHack()
  -- TODO We should somehow signal to the actor generator how much open
  -- space we have; I think it uses that to decide how many monsters to
  -- generate. [cf. src/wild.c lines 350-373]
  return ux, uy, dx, dy, spots
end

-- [cf. wilderness_gen() in src/wild.c]
function _M:genTerrain()
  local Map = require 'mod.class.Map'

  local mw, mh = self.level.map.w, self.level.map.h
  local mx, my = game.player.wild_x, game.player.wild_y
  local lv = self:generateGrids(mx, my, false, false, false)
  local lvv

  -- We generate borders by generating the entire terrain levels for
  -- adjacent world map grids and scraping off their edges:
  --   = North
  lvv = self:generateGrids(mx, my - 1, true, false, true)
  lv[0] = lv[0] or {}
  for x = 1, mw - 2 do lv[0][x] = lvv[mh-2][x] end
  --   = South
  lvv = self:generateGrids(mx, my + 1, true, false, true)
  lv[mh-1] = lv[mh-1] or {}
  for x = 1, mw - 2 do lv[mh-1][x] = lvv[1][x] end
  --   = West
  lvv = self:generateGrids(mx - 1, my, true, false, true)
  for y = 1, mh - 2 do lv[y][0] = lvv[y][mw-2] end
  --   = East
  lvv = self:generateGrids(mx + 1, my, true, false, true)
  for y = 1, mh - 2 do lv[y][mw-1] = lvv[y][1] end
  -- And just for overkill, we do the same with the border corners:
  --   = Northwest
  lvv = self:generateGrids(mx - 1, my - 1, false, true, true)
  lv[0][0] = lv[mh-2][mw-2]
  --   = Northeast
  lvv = self:generateGrids(mx + 1, my - 1, false, true, true)
  lv[0][mw-1] = lv[mh-2][1]
  --   = Southwest
  lvv = self:generateGrids(mx - 1, my + 1, false, true, true)
  lv[mh-1][0] = lv[mh-2][1]
  --   = Southeast
  lvv = self:generateGrids(mx + 1, my + 1, false, true, true)
  lv[mh-1][mw-1] = lv[1][1]

  -- Finally, install terrain on the map according to the grid types in our
  -- array.
  local g = game.state:getWorldGrid(mx, my)
  for y, row in pairs(lv) do
    for x, grid in pairs(row) do
      self.level.map(x, y, Map.TERRAIN, self.grid_list[grid])
    end
  end
end

-- [cf. perturb_point_mid() in src/wild.c]
local function perturb_point_mid(x1, x2, x3, x4, rough, max_depth)
  local avg = math.floor((x1 + x2 + x3 + x4)/4) + rng.range(-rough, rough)
  -- Fake rounding half-up.
  if (x1 + x2 + x3 + x4) % 4 > 1 then avg = avg + 1 end
  return util.bound(avg, 0, max_depth)
end

-- [cf. perturb_point_end() in src/wild.c]
local function perturb_point_end(x1, x2, x3, rough, max_depth)
  local avg = math.floor((x1 + x2 + x3)/3) + rng.range(-rough, rough)
  -- Fake rounding half-up.
  if (x1 + x2 + x3) % 3 > 0 then avg = avg + 1 end
  return util.bound(avg, 0, max_depth)
end

-- [cf. plasma_recursive() in src/wild.c]
local function plasma_recursive(lv, x1, y1, x2, y2, max_depth, rough)
  local xmid = math.floor((x2 + x1)/2)
  local ymid = math.floor((y2 + y1)/2)

  -- Are we done?
  if x1 + 1 >= x2 then return end

  lv[ymid][xmid] = perturb_point_mid(lv[y1][x1], lv[y2][x1], lv[y1][x2], lv[y2][x2], rough, max_depth)
  lv[y1][xmid] = perturb_point_end(lv[y1][x1], lv[y1][x2], lv[ymid][xmid], rough, max_depth)
  lv[ymid][x2] = perturb_point_end(lv[y1][x2], lv[y2][x2], lv[ymid][xmid], rough, max_depth)
  lv[y2][xmid] = perturb_point_end(lv[y2][x2], lv[y2][x1], lv[ymid][xmid], rough, max_depth)
  lv[ymid][x1] = perturb_point_end(lv[y2][x1], lv[y1][x1], lv[ymid][xmid], rough, max_depth)

  -- Recurse to the four quadrants
  plasma_recursive(lv, x1, y1, xmid, ymid, max_depth, rough)
  plasma_recursive(lv, xmid, y1, x2, ymid, max_depth, rough)
  plasma_recursive(lv, x1, ymid, xmid, y2, max_depth, rough)
  plasma_recursive(lv, xmid, ymid, x2, y2, max_depth, rough)
end

-- [cf. generate_area() in src/wild.c]
function _M:generateGrids(mx, my, border, corner, use_map)
  local mw, mh = self.level.map.w, self.level.map.h
  -- World map grids are hard-coded with 18 terrain types [cf. MAX_WILD_TERRAIN
  -- in src/types.h], corresponding to "level" values 0-17 in lv{} [the T2
  -- source code reuses the level 'feat' field for this; I think we can
  -- afford to make our own array].
  local lv = {}
  for y = 0, mh - 1 do lv[y] = {} end
  if not corner then
    for y = 0, mh - 1 do
      for x = 0, mw - 1 do
	lv[y][x] = 9	-- MAX_WILD_TERRAIN/2
      end
    end
  end

  -- Use a fixed RNG seed to effect pseudo-persistence of terrain layout.
  rng.seed(game.state:getWorldMapSeed(mx, my))

  -- Initialize the four corners.
  lv[1][1] = rng.range(0, 17)
  lv[mh-2][1] = rng.range(0, 17)
  lv[1][mw-2] = rng.range(0, 17)
  lv[mh-2][mw-2] = rng.range(0, 17)

  if not corner then
    -- 'roughness' is hard-coded to 1 in generate_area().
    plasma_recursive(lv, 1, 1, mw - 2, mh - 2, 17, 1)
  end

  -- Done with fixed-RNG stuff; go back to random.
  rng.seed(-1)

  -- Convert levels to grid types according to our corresponding world map
  -- grid definition.
  local g = game.state:getWorldGrid(mx, my)
  for y, row in pairs(lv) do
    for x, l in pairs(row) do
      lv[y][x] = g.zone_features[l+1]
    end
  end

  if use_map and g.terrain_type == 'TOWN' and g.change_zone then
    -- If this wilderness zone is a town, overlay the town's grid types as
    -- per its map.  This is important if we're going to be used as the
    -- border of an adjacent wilderness zone.
    local mapfile = g.change_zone:gsub('^town%-', '/data/maps/towns/')..'.lua'
    local f, err = loadfile(mapfile)
    if not f then
      if err then error(err) end
      return lv
    end
    local z = {}
    local mt = {
      defineTile = function(char, grid, _, _, _, _, _) z[char] = grid end
    }
    setfenv(f, setmetatable(mt, {__index=_G}))
    local ret, err = f()
    if not ret then
      if err then error(err) end
      return lv
    end
    for y, row in ipairs(ret:split('\n')) do
      local x = 1
      for c in row:gmatch('.') do
	if z[c] then
	  -- Don't forget, map coords are 0-based.
	  lv[y-1][x-1] = z[c]
	end
	x = x + 1
      end
    end
  end

  return lv
end

function _M:borderHack()
  local mw, mh = self.level.map.w, self.level.map.h
  for x = 0, mw - 1 do
    self.level.map.attrs(x, 0, 'wild_border', true)
    self.level.map.attrs(x, mh-1, 'wild_border', true)
  end
  for y = 1, mh - 2 do
    self.level.map.attrs(0, y, 'wild_border', true)
    self.level.map.attrs(mw-1, y, 'wild_border', true)
  end
end
