-- $Id: PlayerCombat.lua 3187 2017-11-22 22:49:21Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2012-2017 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

require 'engine.class'
local Dialog = require 'engine.ui.Dialog'
local DamageType = require 'engine.DamageType'

module(..., package.seeall, class.make)

_M.melee_styles = {}

function _M:loadMeleeStyles(file)
  local env = {
    PlayerCombat = self,
    newMeleeStyle = function(t) self:newMeleeStyle(t) end,
  }
  local f, err = util.loadfilemods(file, setmetatable(env, {__index = _G}))
  if not f and err then error(err) end
  f()
end

function _M:newMeleeStyle(t)
  assert(t.id, 'no melee style id')
  assert(t.desc, 'no melee style desc')
  assert(t.skill, 'no melee style skill')
  assert(t.one_blow, 'no melee style one_blow() function')
  assert(t.num_blows, 'no melee style num_blows() function')
  assert(t.all_weapons, 'no melee style all_weapons() function')

  t.can_use = t.can_use or function(def, who) return true end
  t.on_select = t.on_select or function(def, who) end
  t.on_deselect = t.on_deselect or function(def, who) return nil end
  self.melee_styles[#self.melee_styles+1] = t
  self.melee_styles[t.id] = t
end

function _M:init(t)
  self.melee_style = t.melee_style or 'WEAPON_COMBAT'
  _M.melee_styles[self.melee_style]:on_select(self)
end

-- cf. critical_norm() in src/cmd1.c
function _M:criticalHit(wpn, dam, skill)
  if not wpn.weight then
    -- The 1d1 barehanded poke in 'Weapon combat' mode with no weapon
    -- wielded doesn't get a critical hit.  I can tell you're disappointed.
    -- ;)
    return dam, false
  end

  local done_crit = false

  -- NOTE:  T2 stored weapon weight in tenths of a pound; we store in
  -- pounds.
  local i = wpn.weight*10 + (self.to_h + wpn.to_h)*5 +
	    self:getSkillLevel(skill, 150) + 50*self.extra_crit +
	    self:getLuckScale(-100, 100)
  if wpn.subtype == 'sword' and wpn.weight < 5 then
    i = i + self:getSkillLevel('CRITICAL_HITS', 2000)
  end
  local eff = self:hasEffect(self.EFF_DEADLY_AIM)
  if eff then
    -- Deadly Aim = guaranteed crit.
    game.logSeen(self, 'It was a *GREAT* hit!')
    dam = 3*dam + 20
    done_crit = true
    eff.dur = eff.dur - 1
    if eff.dur <= 0 then self:removeEffect(self.EFF_DEADLY_AIM) end
  elseif rng.range(1, 5000) <= i then
    -- Success!  Determine damage increase based on weapon weight.
    done_crit = true
    local k = wpn.weight*10 + rng.range(1, 650)
    if wpn.subtype == 'sword' and wpn.weight < 5 then
      k = k + self:getSkillLevel('CRITICAL_HITS', 400)
    end
    if k < 400 then
      game.logSeen(self, 'It was a good hit!')
      dam = 2*dam + 5
    elseif k < 700 then
      game.logSeen(self, 'It was a great hit!')
      dam = 2*dam + 10
    elseif k < 900 then
      game.logSeen(self, 'It was a superb hit!')
      dam = 3*dam + 15
    elseif k < 1300 then
      game.logSeen(self, 'It was a *GREAT* hit!')
      dam = 3*dam + 20
    else
      game.logSeen(self, 'It was a *SUPERB* hit!')
      dam = math.floor((7*dam)/2) + 25
    end
  end

  return dam, done_crit
end

function _M.oneBlowWithWeapon(src, target, wpn)
  local destroy_wpn = false	-- Weapon can be destroyed by Nazgul.

  if wpn:hasFlag('flags', 'NEVER_BLOW') then
    game.logPlayer(src, 'Your %s will not attack!', wpn:getName{no_count=true})
    return false
  end

  local combat = wpn.combat or { 1, 1 }
  local base = rng.dice(combat[1], combat[2])
  local bonus = (wpn.to_d or 0) + (src.to_d or 0) + (src.to_d_melee or 0)
  local dt = DamageType:get(DamageType.PHYSICAL)

  -- Should now be covered as flag callbacks.
  --[[
  if wpn:hasFlag('flags', 'CHAOTIC') and rng.chance(2) then
    -- TODO implement chaotic effects (cf. src/cmd1.c, lines 2688-2715)
  end
  if wpn:hasFlag('flags', 'VAMPIRIC') then
    -- TODO implement vampiric effects (cf. src/cmd1.c, lines 2718-2726)
  end
  ]]

  -- This is handled in :compute_brand_slay_mult() above now.
  --[[
  local mult = 1
  for f, _ in pairs(wpn.slay_flags or {}) do
    local ff = f:gsub('^X_', '')
    local has = target:has(ff)
    if has and target.learnFlags then target:learnFlags('flags', ff) end
    local sm = not has and 1 or ff ~= f and 5 or (ff == 'ANIMAL' or ff == 'EVIL') and 2 or 3
    mult = math.max(mult, sm)
  end
  for dt, _ in pairs(wpn.brand_dam or {}) do
    local vuln = target:vulnDam(dt, {learn=true})
    local res = target:resistDam(dt, {learn=true, or_immune=true})
    mult = math.max(mult, (vuln and 6 or not res and 3 or 1))
    if dt == DamageType.POISON and (vuln or not res) then
      -- Special poison effects (cf. src/cmd1.c lines 476-491)
      local doit = vuln and rng.percent(95) or (rng.percent(50) and not rng.percent(target.level))
      if doit then
	target:setEffect(target.EFF_POISON, vuln and base*2 or base,
			 {merge='combine', combine_msg='#Target# is more poisoned.'})
      end
    end
  end
  base = base * mult
  ]]
  base = base * src:compute_brand_slay_mult(target, wpn)

  -- Covered as a flag callback now.
  --[[
  if wpn:hasFlag('flags', 'WOUNDING') then
    -- Special cut effect (cf. src/cmd1.c lines 495-511)
    if target:has('NO_CUT') and target.learnFlags then
      target:learnFlags('flags', 'NO_CUT')
    elseif rng.percent(50) and not rng.percent(target.level) then
      target:setEffect(target.EFF_CUT, base*2,
		       {merge='combine', combine_msg='#Target# is bleeding more strongly.'})
    end
  end
  ]]

  -- TODO implement backstab
  -- TODO earthquakes from the IMPACT flag (is now flag callback)

  -- Clear this for use by callbacks.
  obj_util.once_per_wpn_hit = {}

  local p = {}
  src:fireCallbacks('on_wpn_hit_pre', target, wpn, p)
  if p.stop then return false end

  local got_crit
  base, got_crit = src:criticalHit(wpn, base, 'WEAPONMASTERY')

  if wpn.subtype == 'hafted' and wpn.weight > 5 and got_crit then
    if rng.percent(src:getSkillLevel('STUNNING_BLOWS')) then
      -- Stun the target (cf. src/cmd1.c lines 2767-2785)
      local res = target:resistDam(DamageType.SOUND, {learn=true})
      res = res or target:knowTalent(who.T_NPC_BR_WALL)
      if not res then
	local eff = target:hasEffect(target.EFF_STUN)
	local dur = src:getSkillLevel('STUNNING_BLOWS', 30) + 10
	if eff then dur = dur*2 end
	target:setEffect(target.EFF_STUN, dur,
			 {merge='combine', combine_msg='@Target@ is more dazed.'})
	eff = target:hasEffect(target.EFF_STUN)
	-- We apparently cap stun duration at 200 turns.
	eff.dur = math.min(eff.dur, 200)
      end
    end
  end

  p = { base=base, bonus=bonus }
  src:fireCallbacks('on_wpn_hit_dam_mod', target, wpn, p)
  base, bonus = p.base, p.bonus

  --[[ Should be covered as a flag callback now.
  if wpn:hasFlag('flags', 'VORPAL') and rng.chance(6) then
    -- Vorpal effects (cf. src/cmd1.c lines 2790-2798)
    local rep = base
    game.logPlayer(self, 'Your weapon cuts deep into %s!', target:getName())
    repeat base = base + rep until not rng.chance(4)
  end
  ]]

  -- Now covered in an actor callback in the pray-to-Tulkas talent.
  --[[
  if src:isPrayingTo('Tulkas') and wpn.type == 'weapon' then
    local dd = (wpn.to_d or 0) + (src.to_d_melee or 0)
    if rng.percent(src:getStat('WIS', {scale=130}) - (target.level or 0)) and src.piety > 1000 then
      game.logPlayer(src, 'You feel the hand of Tulkas helping your blow.')
      bonus = bonus + dd*2
    else
      bonus = bonus + dd
    end
  else
    bonus = bonus + (wpn.to_d or 0)
  end
  ]]

  -- Nazgul effects. [cf. do_nazgul() in src/cmd1.c]
  if wpn.type == 'weapon' and target:has('NAZGUL') then
    local wpn_name = wpn:getName{no_count=true, no_mods=true, no_pval=true, no_inscr=true}
    local no_shatter = wpn:hasFlag('flags', 'RES_MORGUL')
    local mundane = not wpn:hasAnyFlag('slay_flags', 'EVIL', 'UNDEAD', 'X_UNDEAD')

    if not wpn:isArtifact() and not wpn:isEgo() and not no_shatter then
      game.logPlayer(src, 'Your %s #{bold}#*DISINTEGRATES*#{normal}#!', wpn_name)
      base = 0
      destroy_wpn = true
    elseif wpn:isEgo() then
      if mundane then
	game.logPlayer(src, '%s is IMPERVIOUS to your mundane %s.', target:getName(), wpn_name)
	base = 0
      end
      -- 25% chance to destroy ego weapons.
      if rng.percent(25) and not no_shatter then
	game.logPlayer(src, 'Your %s is destroyed!', wpn_name)
	destroy_wpn = true
      end
    elseif wpn:isArtifact() then
      if mundane then
	game.logPlayer(src, '%s is IMPERVIOUS to your mundane %s.', target:getName(), wpn_name)
	base = 0
      end
      spell.disenchant(wpn, src)
      -- 0.1% chance to destroy artifact.
      if rng.chance(1000) and not no_shatter then
	game.logPlayer(src, 'Your %s is destroyed!', wpn_name)
	destroy_wpn = true
      end
    end

    -- Destroy the weapon if necessary.
    if destroy_wpn then
      local _, item, inv_id = src:findInAllInventoriesByObject(wpn)
      if inv_id and item then
	src:removeObject(src:getInven(inv_id), item)
	src:afterTakeoff(wpn)
	src:recalcEverything()
	src.changed = true
      end
    end
    -- If any damage is done, 25% chance of getting Black Breath.
    if base > 0 then
      game.logPlayer(src, 'Your foe calls upon your soul!')
      src:setEffect(src.EFF_BLACK_BREATH, 1, {})
    end
  end

  -- Now covered in an actor callback in the pray-to-Melkor talent.
  --[[
  if src:isPrayingTo('Melkor') then
    -- Melkor Curse spell (cf. src/cmd1.c line 2828)
    local t = src:getTalentFromId(src.T_MELKOR_CURSE)
    local lev = spell.get_spell_level(src, t, 100)
    if lev >= 10 then
      local wis = src:getStat('WIS', {scale=30})
      local pr = math.floor((wis * lev)/target.level)
      pr = math.max(pr, 1)
      if src.piety > 5000 and rng.percent(pr) then
	t.do_curse(src, t, target)
      end
    end
  end
  ]]

  -- TODO cloning effect (should be covered in a flag callback now)

  local dam = { math.max(base + bonus, 0), weapon=wpn }
  dt.projector(src, target.x, target.y, DamageType.PHYSICAL, dam)
  return not destroy_wpn
end

function _M.dmgDescWeapon(ms, src)
  local wpns = ms.all_weapons(src)
  local n_blows = ms.num_blows(src, wpns[1])
  local to_d = src.dis_to_d + src.to_d_melee
  if wpns[1].isIDd and wpns[1]:isIDd() then
    to_d = to_d + (wpns[1].to_d or 0)
  end
  if wpns[1].combat[1] == 0 or wpns[1].combat[2] == 0 then
    return to_d <= 0 and 'nil!' or ('%d'):format(to_d)
  end
  local ret = ('%dd%d'):format(n_blows*wpns[1].combat[1], wpns[1].combat[2])
  if to_d ~= 0 then ret = ret .. ('%+d'):format(n_blows*to_d) end
  return ret
end

local wpn_blows = {
  { 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 3 },
  { 1, 1, 1, 1, 2, 2, 3, 3, 3, 4, 4, 4 },
  { 1, 1, 2, 2, 3, 3, 4, 4, 4, 5, 5, 5 },
  { 1, 2, 2, 3, 3, 4, 4, 4, 5, 5, 5, 5 },
  { 1, 2, 2, 3, 3, 4, 4, 5, 5, 5, 5, 5 },
  { 2, 2, 3, 3, 4, 4, 5, 5, 5, 5, 5, 6 },
  { 2, 2, 3, 3, 4, 4, 5, 5, 5, 5, 5, 6 },
  { 2, 3, 3, 4, 4, 4, 5, 5, 5, 5, 5, 6 },
  { 3, 3, 3, 4, 4, 4, 5, 5, 5, 5, 6, 6 },
  { 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 6, 6 },
  { 3, 3, 4, 4, 4, 4, 5, 5, 5, 6, 6, 6 },
  { 3, 3, 4, 4, 4, 4, 5, 5, 6, 6, 6, 6 },
}
function _M.numBlowsWithWeapon(src, wpn)
  -- Dummy weapon has no weight, and gets one blow.
  if not wpn.weight then return 1 end

  local str = src:getStat('STR')
  local dex = src:getStat('DEX')
  local div = math.max(wpn.weight, src.weapon_blows.base_wgt)*10
  local mul = src.weapon_blows.wgt_mult
  local s_idx = math.floor(((str <= 17 and str or (str - 16)*10)*mul)/div)
  s_idx = util.bound(s_idx, 0, 11)
  local d_idx = dex <= 9  and 0 or
		dex <= 18 and 1 or
		dex <= 22 and 2 or
		dex <= 26 and math.floor((dex - 23)/2) or
		dex <= 34 and dex - 24 or
		dex <= 38 and (dex - 19)*2 or 20
  d_idx = util.bound(d_idx, 0, 11)
  local max = 4 + (src.abilities.EXTRA_MAX_BLOW_1 and 1 or 0) +
		  (src.abilities.EXTRA_MAX_BLOW_2 and 1 or 0)
  return math.min(max, wpn_blows[s_idx+1][d_idx+1])
end

function _M.oneBlowSpecial(src, target, skill, dummy_wpn, blows)
  local sk_lev = src:getSkillLevel(skill)
  local dt = DamageType:get(DamageType.PHYSICAL)

  local ok_blows = {}
  -- Collect the blows we can do.
  for _, blow in ipairs(blows) do
    if blow.min_level <= sk_lev then table.insert(ok_blows, blow) end
  end
  local blow = nil
  if #ok_blows == 0 then
    blow = blows[1]
  else
    -- Try a few times and keep the best blow type we get.
    for _ = 1, math.max(1, math.floor(sk_lev/7)) do
      local try
      -- Higher-level blows are less likely.
      repeat try = rng.table(ok_blows) until rng.range(1, sk_lev) >= try.chance
      if not blow or (try.min_level > blow.min_level and not src:hasEffect(src.EFF_CONFUSED) and not src:hasEffect(src.EFF_STUN)) then
	blow = try
      end
    end
  end

  local dam = rng.dice(blow.dam[1], blow.dam[2]) + (dummy_wpn.to_d or 0)
  dam = dam + (src.to_d or 0) + (src.to_d_melee or 0)
  dam = math.max(0, dam)
  dam = { dam, atk_msg=blow.msg }

  -- TODO apply effects

  local got_crit
  local fake_wpn = {
    weight = (src.level * rng.range(1, 10))/10,
    to_h = blow.min_level,
    weapon = dummy_wpn,
  }
  dam[1], got_crit = src:criticalHit(fake_wpn, dam[1], skill)

  dt.projector(src, target.x, target.y, DamageType.PHYSICAL, dam)
  return true
end

function _M.dmgDescSpecial(ms, src, skill, blows)
  local n_blows = ms.num_blows(src)
  local sk = src:getSkillLevel(skill)
  local min_atk, max_atk
  min_atk = blows[1]
  for _, atk in blows do if atk.min_level <= sk then max_atk = atk end end
  local to_d = src.dis_to_d + src.to_d_melee
  local low = math.max(0, min_atk.dam[1] + to_d)
  local hi = math.max(0, max_atk.dam[1]*max_atk.dam[2] + to_d)
  if low <= 0 and hi <= 0 then return '0' end
  return ('%d-%d'):format(n_blows * low, n_blows * hi)
end

local barehand_blows = {
  {
    msg = " punches %s",
    min_level = 1,
    chance = 0,
    dam = { 2, 4 },
    effects = {},
  },
  {
    msg = " kicks %s",
    min_level = 2,
    chance = 0,
    dam = { 2, 6 },
    effects = {},
  },
  {
    msg = " strikes %s",
    min_level = 3,
    chance = 0,
    dam = { 2, 7 },
    effects = {},
  },
  {
    msg = " hits %s with a knee",
    min_level = 5,
    chance = 5,
    dam = { 4, 3 },
    effects = { KNEE=0 },
  },
  {
    msg = " hits %s with an elbow",
    min_level = 7,
    chance = 5,
    dam = { 2, 8 },
    effects = {},
  },
  {
    msg = " butts %s",
    min_level = 9,
    chance = 10,
    dam = { 4, 5 },
    effects = {},
  },
  {
    msg = " kicks %s",
    min_level = 11,
    chance = 10,
    dam = { 6, 4 },
    effects = { SLOW=0 },
  },
  {
    msg = " uppercuts %s",
    min_level = 13,
    chance = 12,
    dam = { 8, 4 },
    effects = { STUN=6 },
  },
  {
    msg = " double-kicks %s",
    min_level = 16,
    chance = 15,
    dam = { 10, 4 },
    effects = { STUN=8 },
  },
  {
    msg = " hits %s with a Cat's Claw",
    min_level = 20,
    chance = 20,
    dam = { 10, 5 },
    effects = {},
  },
  {
    msg = " hits %s with a jump kick",
    min_level = 25,
    chance = 25,
    dam = { 10, 6 },
    effects = { STUN=10 },
  },
  {
    msg = " hits %s with an Eagle's Claw",
    min_level = 29,
    chance = 25,
    dam = { 12, 6 },
    effects = {},
  },
  {
    msg = " hits %s with a circle kick",
    min_level = 33,
    chance = 30,
    dam = { 12, 8 },
    effects = { STUN=10 },
  },
  {
    msg = " hits %s with an Iron Fist",
    min_level = 37,
    chance = 35,
    dam = { 16, 8 },
    effects = { STUN=10 },
  },
  {
    msg = " hits %s with a flying kick",
    min_level = 41,
    chance = 35,
    dam = { 16, 10 },
    effects = { STUN=12 },
  },
  {
    msg = " hits %s with a Dragon Fist",
    min_level = 45,
    chance = 35,
    dam = { 20, 10 },
    effects = { STUN=16 },
  },
  {
    msg = " hits %s with a Crushing Blow",
    min_level = 48,
    chance = 35,
    dam = { 20, 12 },
    effects = { STUN=18 },
  },
}
function _M.oneBlowBarehand(src, target, dummy_wpn)
  return _M.oneBlowSpecial(src, target, 'BAREHAND_COMBAT', dummy_wpn, barehand_blows)
end

function _M:dmgDescBarehand(src)
  return _M.dmgDescSpecial(self, src, 'BAREHAND_COMBAT', barehand_blows)
end

local bearform_blows = {
  {
    msg = " claws %s",
    min_level = 1,
    chance = 0,
    dam = { 3, 4 },
    effects = { STUN=4 },
  },
  {
    msg = " swats %s",
    min_level = 4,
    chance = 0,
    dam = { 4, 4 },
    effects = { WOUND=20 },
  },
  {
    msg = " bites %s",
    min_level = 9,
    chance = 2,
    dam = { 4, 4 },
    effects = { WOUND=30 },
  },
  {
    msg = " hugs %s",
    min_level = 15,
    chance = 5,
    dam = { 6, 4 },
    effects = { FULL_SLOW=0 },
  },
  {
    msg = " swats and rakes %s",
    min_level = 25,
    chance = 10,
    dam = { 6, 5 },
    effects = { STUN=10, WOUND=10 },
  },
  {
    msg = " hugs and claws %s",
    min_level = 30,
    chance = 15,
    dam = { 6, 6 },
    effects = { FULL_SLOW=60, WOUND=60 },
  },
  {
    msg = " double swats %s",
    min_level = 35,
    chance = 20,
    dam = { 9, 7 },
    effects = { STUN=20, WOUND=20 },
  },
  {
    msg = " double swats and rakes %s",
    min_level = 40,
    chance = 25,
    dam = { 10, 10 },
    effects = { STUN=25, WOUND=25 },
  },
}
function _M.oneBlowBearform(src, target, dummy_wpn)
  return _M.oneBlowSpecial(src, target, 'BEARFORM_COMBAT', dummy_wpn, bearform_blows)
end

function _M:dmgDescBearform(src)
  return _M.dmgDescSpecial(self, src, 'BEARFORM_COMBAT', bearform_blows)
end

local armor_slots = { 'BODY', 'CLOAK', 'OFF_HAND', 'GLOVES', 'BOOTS', 'HELM' }
function _M:armorWeight()
  local total = 0
  for _, slot in ipairs(armor_slots) do
    for _, o in ipairs(self.inven[slot] or {}) do
      total = total + o.weight
    end
  end
  return total
end

function _M:getMeleeStyle()
  return _M.melee_styles[self.melee_style]
end

function _M:setMeleeStyle(style, no_energy)
  if style == self.melee_style then return end

  local cur_def = _M.melee_styles[self.melee_style]
  local new_def = _M.melee_styles[style]
  if not new_def then return end
  if not new_def:can_use(self) then
    -- Shouldn't be possible.
    return
  end

  local err = cur_def:on_deselect(self)
  if err then
    game.logPlayer(self, game.flash.BAD, 'You cannot switch to %s:  %s', new_def.desc, tostring(err))
    return
  end

  self.melee_style = style
  new_def:on_select(self)
  self.changed = true
  self:recalcEverything()
  if not no_energy then self:useEnergy() end
end

function _M:selectMeleeStyle(no_energy)
  local orig_style = self.melee_style
  local styles = {}
  for _, def in ipairs(_M.melee_styles) do
    if def:can_use(self) then
      local name = def.desc
      if def.id == orig_style then name = name .. ' (#LIGHT_BLUE#current#LAST#)' end
      table.insert(styles, { name=name, id=def.id })
    end
  end
  local callback = function(sel)
    if sel and sel.id then self:setMeleeStyle(sel.id, no_energy) end
  end
  ui_util.listPopup('Melee Style', 'Choose style:', styles, 300, callback)
end

function _M:getSkillForMeleeStyle()
  local melee_style = self.melee_style or 'WEAPON_COMBAT'
  return _M.melee_styles[melee_style].skill
end

local skill_by_wpn_subtype = {
  sword = 'SWORD_MASTERY',
  axe = 'AXE_MASTERY',
  hafted = 'HAFTED_MASTERY',
  polearm = 'POLEARM_MASTERY',
}
function _M:getSkillForWeapon()
  local wpns = self:getInven('MAIN_HAND')
  if #wpns == 0 then return nil end
  local sk = nil
  for _, wpn in ipairs(wpns) do
    if sk and sk ~= skill_by_wpn_subtype[wpn.subtype] then return nil end
    sk = skill_by_wpn_subtype[wpn.subtype]
  end
  return sk
end

function _M:attackTarget(target, max_blows)
  local p = { no_attack=false }
  self:fireCallbacks('on_melee_attack_pre', target, p)
  if p.no_attack then return end

  local melee_style = self.melee_style or 'WEAPON_COMBAT'
  local ms = _M.melee_styles[melee_style]
  local wpns = ms.all_weapons(self)

  -- Compute number of blows for each weapon (so in theory, a multi-armed
  -- Possessor wielding Nimthanc and Skullcleaver could get six hits with
  -- the former and four with the latter [which may not be how the original
  -- ToME2 handled it, but...]).
  local n_blows = { total=0 }
  for i, wpn in ipairs(wpns) do
    n_blows[i] = ms.num_blows(self, wpn)
    if max_blows then n_blows[i] = math.min(n_blows[i], max_blows) end
    n_blows.total = n_blows.total + n_blows[i]
  end

  -- Remember this for possible use by Spread Blows.
  local blows, hits = 0, 0

  -- Clear this for callbacks.
  obj_util.once_per_melee_attack = {}

  for i, wpn in ipairs(wpns) do
    if self.dead or target.dead then break end
    for _ = 1, n_blows[i] do
      blows = blows + 1
      local th = self.to_h + self.to_h_melee + (wpn.to_h or 0)
      local chance = self.skill_fighting + th*3
      print(('[COMBAT] PlayerCombat:attackTarget:  %d + 3*(%d + %d + %d) = %d'):format(self.skill_fighting, self.to_h, self.to_h_melee, (wpn.to_h or 0), chance))
      if self:checkHit(target, chance) then
	local keep_going = ms.one_blow(self, target, wpn)
	-- If target is dead (or we're dead from retaliatory damage) or the
	-- one_blow() callback signaled us to stop, stop attacking.
	if not keep_going or target.dead or self.dead then break end
	-- At this point, both attacker and target are alive and we're
	-- still attacking, so we can apply melee blow effects.
	self:fireCallbacks('on_melee_blow', target, true, wpn)
	if target.onHitInMelee then target:onHitInMelee(src, wpn) end
      else
	game.log('%s misses %s.', self.name:capitalize(), target.name)
	self:fireCallbacks('on_melee_blow', target, false, wpn)
      end
    end
  end

  if not self.dead and not target.dead then
    self:fireCallbacks('on_melee_attack', target, hits, blows)
  end

  -- Player speed is apparently factored in by useEnergy().
  local energy = game.energy_to_act
  if melee_style == 'weapon' and self.abilities.SPREAD_BLOWS then
    -- TODO Maybe Spread-Blows-like abilities for barehand and bearform
    -- combat.
    energy = energy * blows/n_blows.total
  end
  self:useEnergy(energy)
end
