-- $Id: store-util.lua 3421 2018-06-14 02:35:56Z dsb $
-- ToME - Tales of Middle-Earth
-- Copyright © 2018 Scott Bigham
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-- Scott Bigham "Zizzo"
-- dsb-tome@killerbunnies.org

-- Assorted store-related utility functions.  Most of these are intended to
-- be called from a chat definition file, sometimes as an action() or
-- cond() field of a newChat{} definition.

store_util = {}

-- Sets a prices{} table in the "global" chat index, which other functions
-- below will expect to find.  The fields of the provided 't' price table
-- should be of the form key={bad_price,normal_price,good_price}.
function store_util.set_prices(owner, player, t)
  prices = {}
  for k, pr in pairs(t) do prices[k] = owner:getRacePrice(player, pr) end
end

-- Returns a function suitable for the cond() method of a newChat{}
-- definition, which returns true if the player has enough gold for the
-- price under the specified key in the global prices{} table.
function store_util.can_afford(key)
  return function(npc, player, dlg)
    return player.gold >= prices[key]
  end
end

-- Returns a function suitable for the cond() method of a newChat{}
-- definition, which returns true if the player does not have enough gold
-- for the price under the specified key in the global prices{} table.
function store_util.cannot_afford(key)
  return function(npc, player, dlg)
    return player.gold < prices[key]
  end
end

-- Returns the provided text with all instance of "@price.<key>@" replaced
-- with the price under the specified key in the global prices{} table.
-- Intended for use with the text field of a newChat{} definition.
function store_util.price_interp(text)
  return text:gsub('@price%.(.-)@', function(key) return prices[key] end)
end

-- Returns a string suffix describing the price under the specified key in
-- the global prices{} table, suitable for appending to the answer text of
-- a newChat{} definition.
function store_util.cost_sfx(key)
  return (' #LIGHT_BLUE#(%d gold)#LAST#'):format(prices[key])
end

-- Common action() method for the listen-for-rumors action in the Inn and
-- the Prancing Pony.  Your chat definition must include a 'rumor' entry,
-- which this method will modify in place.
function store_util.inn_rumor(npc, player, dlg)
  local rumor_pfx = {
    [[The bartender offers you some advice:]],
    [[You overhear someone at the bar say:]],
    [[One of the tables looks rowdy; someone shouts:]],
  }
  local f, err = loadfile('/data/rumors.lua')
  local text
  if f then
    local rumor = f()
    local pfx = rumor_pfx[rng.range(1, #rumor_pfx)]
    text = pfx..'\n\n'..rumor
  else
    print('[RUMOR] failed to load rumors.lua', err)
    text = 'The bar seems quiet; not much to be heard.'
  end
  -- Deep wizardry:  we reach ahead into the next chat object and
  -- replace its text with ours.
  dlg.chat:get('rumor').text = text
  return 'rumor'
end

-- Common action() method for the meal action in the Inn and the Prancing
-- Pony.  Expects the price to be in the chat global prices.inn_meal.
function store_util.inn_meal(npc, player)
  local vampire = player.descriptor.subrace == 'Vampire'
  if not vampire then
    game.log('The barkeep gives you some gruel and a beer.')
    player:adjGold(-prices.inn_meal)
    player:setFood(util.FOOD.max - 1)
  else
    game.log([[You're a vampire; we have no food you can eat.]])
    -- T2 charges the player here anyway, which seems rude.
  end
end

-- Common action() for the room-for-the-night in the Inn and the Prancing
-- Pony.  Expects the price to be in the chat global prices.inn_room.
function store_util.inn_room(npc, player)
  local vampire = player.descriptor.subrace == 'Vampire'
  local hour, _ = game.calendar:getTimeOfDay(game.turn)
  local nighttime = hour < 6 or hour >= 18
  if not vampire and not nighttime then
    game.log('The rooms are available only at night.')
  elseif vampire and nighttime then
    game.log('The rooms are available only during daylight for Vampires.')
  elseif player:has('POISON') or player:has('CUT') then
    game.log([[You need a healer, not a room.  Sorry, but I don't want anyone dying in here.]])
  else
    -- TODO
    local Dialog = require 'engine.ui.Dialog'
    Dialog:simplePopup('Not implemented', 'This feature is not yet implemented.  We Apologize for the Inconvenience.™')
  end
end

-- Common utility method for enchanting bows and arrows in the Rangers
-- Guild and armor and weapons in the Fighters Hall and Craftsmaster.
-- [cf. fix_item in src/bldg.c]
function store_util.enchant_item(player, dlg, filter, names, price, next_chat, shopkeeper_name)
  local max_enchant = math.floor(player.level/5)
  local lines, cands = {}, {}

  local function cb(inven, _, o)
    if not filter(inven, o) then return end
    local name = o:getName{no_mods=true, no_pval=true, no_inscr=true}
    if o:isArtifact() then
      lines[#lines+1] = name .. ':  ' .. (o:isIDd() and 'beyond my skills' or 'in fine condition')
    elseif o.type == 'armor' then
      if o.to_a <= -3 then
	lines[#lines+1] = name .. ':  beyond repair'
      elseif o.to_a >= max_enchant then
	lines[#lines+1] = name .. ':  cannot improve'
      else
	lines[#lines+1] = name .. (':  polish [%d] -> [%d]'):format(o.to_a, o.to_a+1)
	cands[#cands+1] = o
      end
    else
      if o.to_h <= -3 or o.to_d <= -3 then
	lines[#lines+1] = name .. ':  beyond repair'
      elseif o.to_h >= max_enchat and o.to_d >= max_enchant then
	lines[#lines+1] = name .. ':  cannot improve'
      else
	local verbs = {
	  ['shooter/bow']	= 'restring',
	  ['shooter/crossbow']	= 'restring',
	  ['shooter/sling']	= 'restring',
	  ['ammo/shots']	= 'polish',
	  ['weapon/hafted']	= 'refurbish',
	}
	local verb = verbs[o.type..'/'..o.subtype] or 'sharpen'
	local new_to_h = math.min(o.to_h+1, max_enchant)
	local new_to_d = math.min(o.to_d+1, max_enchant)
	lines[#lines+1] = name .. (':  %s (%d,%d) -> (%d,%d)'):format(verb, o.to_h, o.to_d, new_to_h, new_to_d)
	cands[#cands+1] = o
      end
    end
  end
  player:inventoryApplyAll(cb)

  local text = ([[#LIGHT_GREEN#*%s eyes you critically up and down.*#LAST#  You're level %d, right?  Based on that, I can improve your %s if %s less than +%d and not too badly damaged.  Let me see your gear... okay, here's my assessment:]]):format(shopkeeper_name, player.level, names[math.min(#lines, 2)], (#lines > 1 and "they're" or "it's"), max_enchant)
  text = text .. '\n  · ' .. table.concat(lines, '\n  · ')

  local chat = dlg.chat:get(next_chat)
  chat.text = text
  if #cands == 0 then
    chat.answers = {
      { 'Oh well.  Sorry to bother you, then. [leave]' }
    }
  else
    chat.answers = {
      {
	"Cool.  Let's do it, then.  #LIGHT_BLUE#("..price..' gold)#LAST#',
	action = function(npc, player)
	  for _, o in ipairs(cands) do
	    if o.type == 'armor' then
	      o.to_a = math.min(o.to_a+1, max_enchant)
	    else
	      o.to_h = math.min(o.to_h+1, max_enchant)
	      o.to_d = math.min(o.to_d+1, max_enchant)
	    end
	  end
	  player:sortInven()
	  player:adjGold(-price)
	end,
      },
      { "That's all you can do?  Not for that price. [leave]" },
    }
  end

  return next_chat
end

-- Common utility method used by the Rangers' Guild, Fighters Hall and
-- Craftsmaster to check whether the player is wearing equipment we can
-- improve..
-- [cf. fix_item in src/bldg.c]
function store_util.has_gear(player, filter)
  return util.findInAllInventoriesBy(player, function(o, _, inven)
    return filter(inven, o)
  end)
end

-- Common action() method for the identify-possessions action in various
-- stores.  Expects the price to be in the chat global prices.id_pack.
function store_util.id_pack(npc, player)
  player:adjGold(-prices.id_pack)
  player:inventoryApplyAll(function(_, _, o) o:identify(true, player) end)
  game.logPlayer(player, 'Your possessions have been identified.')
end

-- Common action() method for the recharge-item action in various stores.
-- Expects the price to be in the chat global prices.recharge.
function store_util.recharge_item(npc, player)
  -- obj_util.recharge() requires a coroutine context.
  local co = coroutine.create(function()
    if obj_util.recharge(player, 80) then
      player:adjGold(-prices.recharge)
    end
  end)
  local ok, ret = coroutine.resume(co)
  if not ok then
    print(debug.traceback(co))
    if ret then error(ret) end
  end
end

-- Common action() method for the healing-prayer action in various stores.
-- Expects the price to be in the chat global prices.heal.
function store_util.healing(npc, player)
  player:heal(200)
  player:removeEffect(player.EFF_POISON)
  player:removeEffect(player.EFF_BLIND)
  player:removeEffect(player.EFF_CONFUSED)
  player:removeEffect(player.EFF_CUT)
  player:removeEffect(player.EFF_STUN)
  player:adjGold(-prices.heal)
end

-- Common action() method for the restoration action in various stores.
-- Expects the price to be in the chat global prices.restore.
function store_util.restoration(npc, player)
  local Stats = require 'mod.class.interface.ActorStats'
  local any = false
  for _, s in ipairs(Stats.stats) do
    if player:restoreStat(s) then any = true end
  end
  if any then
    player:adjGold(-prices.restore)
  end
end

-- Common action() method for the research-item action in various stores.
-- Expects the price to be in the chat global prices.id.
function store_util.full_id_item(npc, player)
  local function filter_cb(o) return not o:isIDd(true) end
  local function action_cb(o, _, _, _, _, _)
    if o then
      player:adjGold(-prices.id)
      o:identify('full', player)
      local Dialog = require 'engine.ui.Dialog'
      Dialog:simpleLongPopup(o:getName(), o:getDesc(), 600, nil, false, 0.8)
      player:sortInven()
      if player.recalcEverything then player:recalcEverything() end
    end
    return true
  end
  local ShowEquipInvenFloor = require 'mod.dialogs.ShowEquipInvenFloor'
  game:registerDialog(ShowEquipInvenFloor.new('Select item to fully identify', player, filter_cb, action_cb))
end

-- Common cond() method for the research-teim action in various stores;
-- returns true if the player has enough gold and at least one un-*ID*'d
-- item.  Expects the price to be in the chat global prices.id.
function store_util.ok_full_id_item(npc, player)
  local function cb(o, _, _) return not o:isIDd(true) end
  return player.gold >= prices.id and util.findInAllInventoriesBy(player, cb)
end
